const std = @import("std");

const util = @import("util.zig");
const shell_ = @import("shell.zig");

const Shell = shell_.Shell;
const StdIO = shell_.StdIO;
const Files = shell_.StdIO.Files;
const Term = std.process.Child.Term;
const CMD_NAME_INVALID = shell_.CMD_NAME_INVALID;
const EXEC_PATH: enum { resolve, keep } = .keep;

fn resolve_exec_path(
    cmd_name: []const u8,
    dir_entry: []const u8,
    allocator: std.mem.Allocator,
) ![]u8 {
    const dir_path = try util.realpathWithTilde(allocator, dir_entry);
    defer allocator.free(dir_path);
    const dir_fd = try std.fs.openDirAbsolute(dir_path, .{});
    const stat = try dir_fd.statFile(cmd_name);
    if (stat.kind != .file) return error.NotAFile;
    const file_path = try std.fs.path.join(allocator, &.{ dir_path, cmd_name });
    errdefer allocator.free(file_path);
    try std.posix.access(file_path, std.posix.X_OK);
    return file_path;
}

pub fn find_exec_path(
    cmd_name: []const u8,
    allocator: std.mem.Allocator,
) !?[]u8 {
    std.debug.assert(std.mem.indexOfAny(u8, cmd_name, CMD_NAME_INVALID) == null);
    const entrylist = try std.process.getEnvVarOwned(allocator, "PATH");
    defer allocator.free(entrylist);
    var iter = std.mem.tokenizeScalar(u8, entrylist, ':');
    while (iter.next()) |entry| {
        const file_path = resolve_exec_path(cmd_name, entry, allocator) catch continue;
        return file_path;
    }
    return null;
}

/// stolen from std.process.Child
fn statusToTerm(status: u32) Term {
    return if (std.posix.W.IFEXITED(status))
        Term{ .Exited = std.posix.W.EXITSTATUS(status) }
    else if (std.posix.W.IFSIGNALED(status))
        Term{ .Signal = std.posix.W.TERMSIG(status) }
    else if (std.posix.W.IFSTOPPED(status))
        Term{ .Stopped = std.posix.W.STOPSIG(status) }
    else
        Term{ .Unknown = status };
}

pub fn run_command(
    argv: []const []const u8,
    files: Files,
    allocator: std.mem.Allocator,
) !u8 {
    if (argv.len == 0) return error.NoArgv;

    // copy env
    const env = try std.process.createEnvironFromExisting(
        allocator,
        @ptrCast(std.os.environ.ptr),
        .{},
    );

    // check that the command exits
    const exec_path = try find_exec_path(argv[0], allocator) orelse return error.NoSuchCommand;
    // FIXME: THIS IS NOT SAFE, since there's space for race condition between
    //        the check and the execvpeZ call below.
    //        I haven't found any better way to check, though: execvpeZ() call below
    //        is running in child process so its error.FileNotFound is not really useful

    // set up argv
    const argv_ = a: {
        var al = std.ArrayList([]const u8).init(allocator);
        try al.append(switch (EXEC_PATH) {
            .keep => argv[0],
            .resolve => exec_path,
        });
        try al.appendSlice(argv[1..]);
        break :a try al.toOwnedSlice();
    };

    // make it into zero-terminated
    const argv_buf = try allocator.allocSentinel(?[*:0]const u8, argv_.len, null);
    for (argv_, 0..) |arg, i| {
        argv_buf[i] = (try allocator.dupeZ(u8, arg)).ptr;
    }

    // fork()!
    const child_pid = try std.posix.fork();
    if (child_pid == 0) {
        // WE ARE THE CHILD

        // copy the file handles
        if (files.stdin) |f| std.posix.dup2(f.handle, std.posix.STDIN_FILENO) catch {};
        if (files.stdout) |f| std.posix.dup2(f.handle, std.posix.STDOUT_FILENO) catch {};
        if (files.stderr) |f| std.posix.dup2(f.handle, std.posix.STDERR_FILENO) catch {};
        const err = std.posix.execvpeZ(argv_buf.ptr[0].?, argv_buf.ptr, env.ptr);
        _ = err catch std.process.exit(shell_.EXIT_ERROR); // nobody should see this, though..
    } else {
        // WE ARE THE PARENT

        // close all the files and wait
        // if (files.stdin) |f| f.close();
        // if (files.stdout) |f| f.close();
        // if (files.stderr) |f| f.close();
        const res = std.posix.waitpid(child_pid, 0);
        const term = statusToTerm(res.status);
        if (term == .Exited) return term.Exited;
        return error.UnsupportedWaitResult;
    }
}
