const std = @import("std");

pub const PwdComputer = struct {
    byenv: []const u8,
    byenv_resolved: []const u8,
    byenv_realpath: []const u8,
    actual: []const u8,
    actual_resolved: []const u8,
    allocator: std.mem.Allocator,

    // POSIX (speaks in POSIXese)
    //
    // About pwd options (<https://pubs.opengroup.org/onlinepubs/9699919799/utilities/pwd.html>):
    //
    // > The following options shall be supported by the implementation:
    // >
    // >  * `-L` - If the PWD environment variable contains an absolute pathname
    // >    of the current directory and the pathname does not contain any components
    // >    that are dot or dot-dot, pwd shall write this pathname to standard output,
    // >    except that if the PWD environment variable is longer than {PATH_MAX} bytes
    // >    including the terminating null, it is unspecified whether pwd writes this
    // >    pathname to standard output or behaves as if the -P option had been specified.
    // >
    // >    Otherwise, the -L option shall behave as the -P option.
    // >
    // > *  `-P` - The pathname written to standard output shall not contain any components
    // >    that refer to files of type symbolic link. If there are multiple pathnames
    // >    that the pwd utility could write to standard output, one beginning with a single
    // >    *slash* character and one or more beginning with two *slash* characters,
    // >    then it shall write the pathname beginning with a single *slash* character.
    // >
    // >    The pathname shall not contain any unnecessary *slash* characters after the
    // >    leading one or two *slash* characters.
    // >
    // > If both -L and -P are specified, the last one shall apply. If neither -L nor -P is
    // > specified, the pwd utility shall behave as if -L had been specified.
    //
    // About PWD (<https://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap08.html>):
    //
    // > This variable shall represent an absolute pathname of the current working directory.
    // > It shall not contain any components that are dot or dot-dot. The value is set by the
    // > cd utility, and by the sh utility during initialization.

    pub fn create(allocator: std.mem.Allocator) !PwdComputer {
        const byenv = std.process.getEnvVarOwned(allocator, "PWD") catch |e| switch (e) {
            error.EnvironmentVariableNotFound => return error.UndefinedBehavior,
            else => return e,
        };
        const actual = try std.process.getCwdAlloc(allocator);
        return PwdComputer{
            .byenv = byenv,
            .byenv_resolved = try std.fs.path.resolvePosix(allocator, &.{byenv}),
            .byenv_realpath = try std.fs.realpathAlloc(allocator, byenv),
            .actual = actual,
            .actual_resolved = try std.fs.path.resolvePosix(allocator, &.{actual}),
            .allocator = allocator,
        };
    }

    pub fn destroy(self: @This()) void {
        self.allocator.free(self.byenv);
        self.allocator.free(self.byenv_resolved);
        self.allocator.free(self.actual);
        self.allocator.free(self.actual_resolved);
    }

    fn _has_dot_or_dotdot(path: []const u8) bool {
        var it = std.mem.tokenizeScalar(u8, path, '/');
        while (it.next()) |component| switch (component.len) {
            0 => continue,
            1 => if (component[0] == '.') return true,
            2 => if (component[0] == '.' and component[1] == '.') return true,
            else => continue,
        };
        return false;
    }

    fn _is_too_long(path: []const u8) bool {
        return (path.len + 1) > std.posix.PATH_MAX;
    }

    fn _l_applies(self: @This()) bool {
        // > If the PWD environment variable contains an absolute pathname
        if (self.byenv.len < 1) return false; // empty
        if (self.byenv[0] != '/') return false; // not absolute
        // > ... of the current directory
        if (!std.mem.eql(u8, self.byenv_resolved, self.actual_resolved)) return false;
        // > ... and the pathname does not contain any components that are dot or dot-dot,
        if (_has_dot_or_dotdot(self.byenv)) return false;
        return true;
    }

    pub fn getpath_P(self: @This()) []const u8 {
        return self.byenv_realpath;
    }

    pub fn getpath_L(self: @This()) []const u8 {
        if (!self._l_applies()) return self.getpath_P();
        // > .. except that if the PWD environment variable is longer than {PATH_MAX} bytes
        // > including the terminating null..
        if ((self.byenv.len + 1) > std.posix.PATH_MAX) {
            // > .. it is unspecified whether pwd writes this pathname to standard output
            // > or behaves as if the -P option had been specified
            //
            // We choose the second. option
            return self.getpath_P();
        }
        return self.byenv;
    }
};
